package processing.app.debug;

@SuppressWarnings("serial")
public class TargetPlatformException extends Exception {

  public TargetPlatformException() {
    super();
  }

  public TargetPlatformException(String arg0, Throwable arg1) {
    super(arg0, arg1);
  }

  public TargetPlatformException(String arg0) {
    super(arg0);
  }

  public TargetPlatformException(Throwable arg0) {
    super(arg0);
  }

}
